package main

import (
	"crypto/hmac"
	"crypto/sha256"
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/gorilla/websocket"
	"io"
	"io/ioutil"
	"net/http"
	"net/url"
	"os"
	"strings"
	"time"
)

var (
	hostUrl   = "ws://ist-api-sg.xf-yun.com/v2/ist"
	apiKey    = "********************************"
	apiSecret = "********************************"
	appid     = "********"
	file      = "********************************"
)

const (
	STATUS_FIRST_FRAME       = 0
	STATUS_CONTINUE_FRAME    = 1
	STATUS_LAST_FRAME        = 2
	HttpCodeSuccessHandshake = 101
)

func main() {
	start()

}

type RespData struct {
	Code    int    `json:"code"`
	Message string `json:"message"`
	Data    Data   `json:"data"`
	Sid     string `json:"sid"`
}

type Data struct {
	Result *Result `json:"result"`
	Status int     `json:"status"`
}

type Decoder struct {
	results []*Result
}

func (d *Decoder) Decode(result *Result) {
	if len(d.results) <= result.Sn {
		d.results = append(d.results, make([]*Result, result.Sn-len(d.results)+1)...)
	}
	if result.Pgs == "rpl" {
		for i := result.Rg[0]; i <= result.Rg[1]; i++ {
			d.results[i] = nil
		}
	}
	d.results[result.Sn] = result
}

func (d *Decoder) String() string {
	var r string
	for _, v := range d.results {
		if v == nil {
			continue
		}
		r += v.String()
	}
	return r
}

type Result struct {
	Ls  bool   `json:"ls"`
	Rg  []int  `json:"rg"`
	Sn  int    `json:"sn"`
	Pgs string `json:"pgs"`
	Ws  []Ws   `json:"ws"`
}

func (t *Result) String() string {
	var wss string
	for _, v := range t.Ws {
		wss += v.String()
	}
	return wss
}

type Ws struct {
	Bg int  `json:"bg"`
	Cw []Cw `json:"cw"`
}

func (w *Ws) String() string {
	var wss string
	for _, v := range w.Cw {
		wss += v.W
	}
	return wss
}

type Cw struct {
	Sc int    `json:"sc"`
	W  string `json:"w"`
}

func readResp(resp *http.Response) string {
	if resp == nil {
		return ""
	}
	b, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		panic(err)
	}
	return fmt.Sprintf("code=%d,body=%s", resp.StatusCode, string(b))
}

func assembleAuthUrl(hosturl string, apiKey, apiSecret string) string {
	ul, err := url.Parse(hosturl)
	if err != nil {
		fmt.Println(err)
	}
	date := time.Now().UTC().Format(time.RFC1123)
	signString := []string{"host: " + ul.Host, "date: " + date, "GET " + ul.Path + " HTTP/1.1"}
	sgin := strings.Join(signString, "\n")
	sha := HmacWithShaTobase64("hmac-sha256", sgin, apiSecret)
	authUrl := fmt.Sprintf("hmac username=\"%s\", algorithm=\"%s\", headers=\"%s\", signature=\"%s\"", apiKey,
		"hmac-sha256", "host date request-line", sha)
	authorization := base64.StdEncoding.EncodeToString([]byte(authUrl))

	v := url.Values{}
	v.Add("host", ul.Host)
	v.Add("date", date)
	v.Add("authorization", authorization)
	callurl := hosturl + "?" + v.Encode()
	return callurl
}

func HmacWithShaTobase64(algorithm, data, key string) string {
	mac := hmac.New(sha256.New, []byte(key))
	mac.Write([]byte(data))
	encodeData := mac.Sum(nil)
	return base64.StdEncoding.EncodeToString(encodeData)
}

type WebsocketListener interface {
	OnMessage(data []byte)
	OnOpen(resp *http.Response)
	OnFail(err error)
	OnClose(err error)
}

type WebSocketClient struct {
	conn     *websocket.Conn
	Listener WebsocketListener
	Url      string
}

func (ws *WebSocketClient) Connect() {
	d := websocket.Dialer{}
	conn, rsp, err := d.Dial(ws.Url, nil)
	if err != nil {
		emsg := ""
		if rsp != nil {
			b, _ := ioutil.ReadAll(rsp.Body)
			emsg = fmt.Sprintf("emsg=%s,responseCode=%d,body=%s", err.Error(), rsp.StatusCode, string(b))
		}
		ws.Listener.OnFail(errors.New(emsg))
		return
	}
	ws.conn = conn
	ws.Listener.OnOpen(rsp)
	go ws.read()
}

func (ws *WebSocketClient) read() {
	for {
		_, msg, err := ws.conn.ReadMessage()
		if err != nil {
			ws.Listener.OnClose(err)
			return
		}
		ws.Listener.OnMessage(msg)
	}
}

func (ws *WebSocketClient) Send(v interface{}) error {
	return ws.conn.WriteJSON(v)
}

func (ws *WebSocketClient) Close() {
	ws.conn.Close()
}

/*
*
OnMessage(data []byte)
OnOpen(resp *http.Response)
OnFail(err error)
OnClose(err error)
*/
type IstWsClient struct {
	Url     string
	Decoder *Decoder
	client  *WebSocketClient
}

func (c *IstWsClient) OnOpen(resp *http.Response) {
	fmt.Println("success open connection")
}

func (c *IstWsClient) OnMessage(data []byte) {
	var resp RespData
	json.Unmarshal(data, &resp)
	if resp.Code != 0 {
		fmt.Println("error|", resp.Code, resp.Message, resp.Sid)
		return
	}
	c.Decoder.Decode(resp.Data.Result)
	fmt.Println("result:", c.Decoder.String(), resp.Sid)
	if resp.Data.Status == 2 {

		fmt.Println("fin====")
		c.client.Close()
	}
}

func (c *IstWsClient) OnFail(err error) {
	fmt.Println(err.Error())
}

func (c *IstWsClient) OnClose(err error) {
	fmt.Println(err.Error())
}

func start() {
	ist := &IstWsClient{
		Decoder: &Decoder{results: make([]*Result, 0)},
	}
	client := &WebSocketClient{
		Listener: ist,
		Url:      assembleAuthUrl(hostUrl, apiKey, apiSecret),
	}
	ist.client = client
	client.Connect()
	audioFile, err := os.Open(file)
	if err != nil {
		panic(err)
	}

	var frameSize = 1280
	var intervel = 40 * time.Millisecond
	var status = STATUS_FIRST_FRAME
	var buffer = make([]byte, frameSize)
	var count = 0
	for {
		var len, err = audioFile.Read(buffer)
		if err != nil {
			if err == io.EOF {
				status = STATUS_LAST_FRAME
			} else {
				panic(err)
			}
		}

		switch status {
		case STATUS_FIRST_FRAME:
			frameData := map[string]interface{}{
				"common": map[string]interface{}{
					"app_id": appid,
				},
				"business": map[string]interface{}{
					"aue":      "raw",
					"language": "en_us",
					"accent":   "mandarin",
					"rate":     "16000",
					"domain":   "ist_open",
					"dwa":      "wpgs",
					"ent":      "ist",
					"eos":      30000000,
				},
				"data": map[string]interface{}{
					"status": STATUS_FIRST_FRAME,
					"audio":  base64.StdEncoding.EncodeToString(buffer[:len]),
				},
			}
			client.Send(frameData)
			status = STATUS_CONTINUE_FRAME
		case STATUS_CONTINUE_FRAME:
			frameData := map[string]interface{}{
				"data": map[string]interface{}{
					"status": STATUS_CONTINUE_FRAME,
					"audio":  base64.StdEncoding.EncodeToString(buffer[:len]),
				},
			}
			client.Send(frameData)
		case STATUS_LAST_FRAME:
			frameData := map[string]interface{}{
				"data": map[string]interface{}{
					"status": STATUS_LAST_FRAME,
					"audio":  base64.StdEncoding.EncodeToString(buffer[:len]),
				},
			}
			client.Send(frameData)
			goto end
		}
		time.Sleep(intervel)
		count++

	}
end:

	time.Sleep(2 * time.Second)
}
